import pandas as pd
from .helper import Process
from ..sources import AidData, SuperAidData
from ..sources.helper import Gas, Price, networks

class AidProcessing: 
    """this class is used to process the data from the aid datasource"""

    def __init__(self, subgrounds, chain_id, AidData):
        """constructor method to initialize the data source class"""

        self.price = Price()
        self.process = Process()
        self.network = networks[chain_id]()
        self.market_aid = AidData #AidData(subgrounds, chain_id)
        
    # TODO: clean this function and make it more efficient
    def build_aid_history(self, aid, bin_size=60):
        """this function serves as an easy way to build back the entire asset history of the aid contract along with price support for the assets.
        it relies heavily upon the marke-aid subgraph as a source of data and applies the necessary transformations to the data to format it in 
        a way that allows for granularity of asset balances and relevant price data to the minute."""

        # get the aid history
        data = self.market_aid.get_aid_history(aid, bin_size)

        # get all of the relevant tokens
        tokens = list(data['aids_balances_token_symbol'].unique())

        # get the min and max timestamps
        min_timestamp = data['aids_balances_history_timestamp'].min()
        max_timestamp = data['aids_balances_history_timestamp'].max()

        # get the tickers needed to retrieve coinbase price data
        tickers = [self.network.coinbase_tickers[token] for token in tokens]

        # group the data and get the cumulative balance changes for each timestamp
        history = data.groupby(['aids_id', 'aids_balances_token_id', 'aids_balances_token_symbol', 'aids_balances_history_time_bin'])
        history = history[['aids_balances_history_balance_change_formatted']].sum() 
        history.reset_index(inplace=True)
        asset_grouping = history.groupby('aids_balances_token_symbol')

        # for each asset, create a seperate dataframe that can be used to get the relevant balance changes for that asset
        asset_changes = {}
        for name, group in asset_grouping:
            asset_changes[name] = dict(zip(group['aids_balances_history_time_bin'], group['aids_balances_history_balance_change_formatted']))

        # collect price data for each asset of interest over the given time period of interest
        price_data = {}

        for token, ticker in zip(tokens, tickers):
            price_data[token] = self.price.get_price_in_range(start = min_timestamp, end = max_timestamp, pair = ticker)

        # create a datarame that is every timestamp (bin_size is the period of interest) and map the relevant price data to each timestamp
        longest_key = max(price_data, key=lambda k: len(price_data[k]))
        price_df = pd.DataFrame.from_dict(price_data[longest_key], orient='index', columns=[f'{longest_key}_price'])

        # reset the index and set the symbol column to the longest key
        price_df.index.name = 'timestamp'
        price_df.reset_index(inplace=True)
        price_df[f'{longest_key}'] = longest_key

        # now go through the tokens and add the price data and balance changes to the dataframe
        for symbol in tokens: 
            if symbol == longest_key:
                price_df[f'{symbol}_balance_change'] = price_df.apply(lambda x: asset_changes[symbol].get(x['timestamp'], 0),  axis=1)
                continue
            price_df[f'{symbol}_price'] = price_df.apply(lambda x: self.process.get_closest_timestamp_value(price_data[symbol], x['timestamp']), axis=1) 
            price_df[f'{symbol}'] = symbol
            price_df[f'{symbol}_balance_change'] = price_df.apply(lambda x: asset_changes[symbol].get(x['timestamp'], 0),  axis=1)

        # sort the dataframe to be in ascending order by timestamp and reset the index
        price_df.sort_values(by=['timestamp'], inplace=True)
        price_df.reset_index(inplace=True, drop=True)

        # now forwad fill the price data, token change data, and compute running balances for each token
        for symbol in tokens: 
            price_df[f'{symbol}_balance'] = price_df[f'{symbol}_balance_change'].cumsum()
            price_df[f'{symbol}_balance_usd'] = price_df[f'{symbol}_balance'] * price_df[f'{symbol}_price']

        # return the dataframe and the tokens in a dictionary 
        return {'data': price_df, 'tokens': tokens}


#from helper import Gas
from .helper import Gas
from subgrounds import Subgrounds
from subgrounds.pagination import ShallowStrategy

from rubi.data.sources.helper import networks

class AidData: 
    """this class acts as the main access point to a variety of data and tooling for MarketAid.sol contracts.
    the MarketAid.sol contracts are meant to be a hub for extended functionality for automated traders on the Rubicon protocol.
    the end goal of this class, and this SDK by extension, is to act as a collaborative platform for the advancement of shared data tooling and analysis.
    """

    def __init__(self, subgrounds, chain_id):
        """constructor method

        :param subgrounds: the subgrounds object
        :type subgrounds: subgrounds.Subgrounds
        :param chain_id: the chain id of the network that is of interest
        :type chain_id: int
        """
        self.network = networks[chain_id]()
        self.subgrounds = subgrounds
        self.market_aid = self.subgrounds.load_subgraph(self.network.market_aid)
    
    # TODO: improvement outlined in issue #19
    def get_aid_history(self, aid=None, bin_size=60, first=1000000000):

        Aid = self.market_aid.Aid
        AidToken = self.market_aid.AidToken
        AidTokenHistory = self.market_aid.AidTokenHistory

        AidToken.balance_formatted = AidToken.balance / 10 ** AidToken.token.decimals
        AidTokenHistory.balance_formatted = AidTokenHistory.balance / 10 ** AidTokenHistory.aid_token.token.decimals
        AidTokenHistory.balance_change_formatted = AidTokenHistory.balance_change / 10 ** AidTokenHistory.aid_token.token.decimals
        AidTokenHistory.time_bin = ((AidTokenHistory.timestamp // bin_size) * bin_size)

        where = []
        if aid:
            where.append(Aid.id == aid)
        
        if where == []:
            aids = self.market_aid.Query.aids(first=first)
        else:
            aids = self.market_aid.Query.aids(first=first, where=where)

        field_paths = [
            aids.id,
            aids.balances.token.id,
            aids.balances.token.symbol,
            aids.balances.history.timestamp,
            aids.balances.history.time_bin,
            aids.balances.history.balance_formatted,
            aids.balances.history.balance_change_formatted,
            aids.balances.history.balance_change
        ]

        df = self.subgrounds.query_df(field_paths, pagination_strategy=ShallowStrategy)
        
        return df

    def get_aid_txns(self, aid=None, start_time=None, end_time=None, first=1000000000): 

        Transaction = self.market_aid.Transaction

        where = []
        if aid:
            where.append(Transaction.aid == aid)
        if start_time:
            where.append(Transaction.timestamp >= start_time)
        if end_time:
            where.append(Transaction.timestamp <= end_time)

        if where == []:
            txns = self.market_aid.Query.transactions(first=first)
        else:
            txns = self.market_aid.Query.transactions(first=first, where=where)

        field_paths = [
            txns.id,
            txns.timestamp,
            txns.aid.id
        ]

        df = self.subgrounds.query_df(field_paths, pagination_strategy=ShallowStrategy)

        return df

class SuperAidData(AidData): 
    """this class acts as an extension of the AidData class and has some additional functionality that is enabled by a node connection. 
    """

    def __init__(self, w3, subgrounds, chain_id):
        """a super class of the AidData class that has additional functionality enabled by a node connection.

        :param w3: a web3 object
        :type w3: web3.main.Web3
        :param subgrounds: the subgrounds object
        :type subgrounds: subgrounds.Subgrounds
        :param chain_id: the chain id of the network that is of interest
        :type chain_id: int
        """
        super().__init__(subgrounds, chain_id)

        # set class variables
        self.w3 = w3
        self.gas = Gas(self.w3)
    
    def get_aid_txns_gas_data(self, aid=None, start_time=None, end_time=None, total_fee_eth=True, total_fee_usd=True, l2_gas_price=None, l2_gas_used=None, l1_gas_used=None, 
        l1_gas_price=None, l1_fee_scalar=None, l1_fee=None, l2_fee=None, total_fee=None, l1_fee_eth=None, l2_fee_eth=None, eth_price=None, l1_fee_usd=None, l2_fee_usd=None, first=1000000000):

        # get the transactions for the market aid contract
        txns = self.get_aid_txns(aid, start_time, end_time, first)
    
        # update the dataframe to contain the relevant gas data
        df = self.gas.txn_dataframe_update(txns, 'transactions_id', 'transactions_timestamp', total_fee_eth = total_fee_eth, total_fee_usd = total_fee_usd, l2_gas_price = l2_gas_price, l2_gas_used = l2_gas_used, l1_gas_used = l1_gas_used, l1_gas_price = l1_gas_price, l1_fee_scalar = l1_fee_scalar, l1_fee = l1_fee, l2_fee = l2_fee, total_fee = total_fee, l1_fee_eth = l1_fee_eth, l2_fee_eth = l2_fee_eth, eth_price = eth_price, l1_fee_usd = l1_fee_usd, l2_fee_usd = l2_fee_usd)

        return df 

    def get_aid_gas_spend_binned(self, aid=None, start_time=None, end_time=None, granularity=60, type = 'USD', total_fee_eth=True, total_fee_usd=True, l2_gas_price=None, l2_gas_used=None, l1_gas_used=None, 
        l1_gas_price=None, l1_fee_scalar=None, l1_fee=None, l2_fee=None, total_fee=None, l1_fee_eth=None, l2_fee_eth=None, eth_price=None, l1_fee_usd=None, l2_fee_usd=None, first=1000000000):
        """this function takes in a market aid address, start time, end time, and granularity in order to return a dictionary of the gas spend dictionary that is binned by the granularity.
        this dictionary is a key-pair mapping of the time bin (start of a minute, hour, day, etc) to the gas spend of the contract during that period.
        """

        if start_time: 
            start_time = int(start_time)
        if end_time:
            end_time = int(end_time)

        gas_spend = self.get_aid_txns_gas_data(aid, start_time, end_time, total_fee_eth, total_fee_usd, l2_gas_price, l2_gas_used, l1_gas_used, l1_gas_price, l1_fee_scalar, l1_fee, l2_fee, total_fee, l1_fee_eth, l2_fee_eth, eth_price, l1_fee_usd, l2_fee_usd, first)

        # bin the data
        gas_spend['time_bin'] = gas_spend['transactions_timestamp'].apply(lambda x: (x // granularity) * granularity)

        # group by time_bin and get the aggregate gas spend during the period represented as a dictionary with time_bin as the key and the gas spend as the value
        grouped = gas_spend.groupby(['time_bin'])

        if type == 'USD':
            gas_spend = grouped['total_fee_usd'].sum().to_dict()

        if type == 'ETH':
            gas_spend = grouped['total_fee_eth'].sum().to_dict()

        return gas_spend




import json
import logging as log
from web3 import Web3
from attributedict.collections import AttributeDict

from rubi.book import Book
import rubi.contracts as contracts
from rubi.contracts.helper import networks
from rubi.data import Data, SuperData

class RubiconData:
    """this class serves to tie in the various data sources and tooling for the Rubicon protocol that can be collected without a node connection. the Rubicon class inherits the functionality of this data class while adding the ability to interact with the contracts. more to come soon!
    """

    def __init__(self, w3=None):
        """initializes the data class, if a web3 instance is provided, it will also initialize the super data class. more to come soon!"""
        if w3:
            self.data = SuperData(w3)
        else:
            self.data = Data()


class Rubicon(RubiconData):
    """this class serves as a the main entry point to the repository. it acts as the initialization of multiple contract instances, and give access to the various functions of the contracts. it also provides a few helper functions to make interacting with the contracts easier. more to come soon!
    
    :param w3: a web3 instance
    :type w3: Web3
    :param wallet: the wallet address of the user, defaults to None
    :type wallet: str, optional
    :param key: the private key of the user, defaults to None
    :type key: str, optional
    :param market: a market contract object, defaults to None
    :type market: str, optional
    :param router: a router contract object, defaults to None
    :type router: str, optional
    :param factory: a factory contract object, defaults to None
    :type factory: str, optional
    """

    def __init__(self, w3, wallet=None, key=None, market=None, router=None, factory=None):
        
        # initialize the data class
        super().__init__(w3)

        # load required variables
        chain = w3.eth.chain_id

        try:
            network = networks[chain]()
        except:
            network = 'unknown'

        # set class variables 
        self.w3 = w3
        self.chain = chain
        self.network = network

        self.key = key
        self.wallet = wallet

        # if wallet and key are provided then create signer instances 
        if wallet and key:
            self.market = contracts.RubiconMarketSigner(w3, wallet, key, contract=market)
            self.router = contracts.RubiconRouterSigner(w3, wallet, key, contract=router)
            self.factory = contracts.FactoryAidSigner(w3, wallet, key, contract=factory)
        else:
            self.market = contracts.RubiconMarket(w3, contract=market)
            self.router = contracts.RubiconRouter(w3, contract=router)
            self.factory = contracts.FactoryAid(w3, contract=factory)

        # set the hashes of the event emits
        self.log_make_hash = self.market.get_log_make_hash()
        self.log_take_hash = self.market.get_log_take_hash()
        self.log_kill_hash = self.market.get_log_kill_hash()
        self.offer_deleted_hash = self.market.get_offer_deleted_hash()
        self.log_swap_hash = self.router.get_log_swap_hash()

        # create a dictionary to map market event emits to their respective functions
        self.market_events = {
            self.log_make_hash : self.market.stream_log_make,
            self.log_take_hash : self.market.stream_log_take,
            self.log_kill_hash : self.market.stream_log_kill
        }

        # create a dictionary to store the market books that are being tracked 
        self.books = {}

    ######################################################################
    # read calls
    ######################################################################

    # a function that will identify the event topic and call the appropriate function
    def parse_market_events(self, event): 
        """a function to parse event emits from the market contract and call the appropriate function to handle the event

        :param event: the event emit from the market contract
        :type event: dict
        :return: the parsed event
        :rtype: dict
        """

        data = AttributeDict(json.loads(event))

        try: 
            parsed = self.market_events[data.params.result.topics[0]](data)
        except:
            return None

        return parsed

    # a function to return all offers for a given token pair, at the depth of the direction specified
    def get_offers(self, token0, token1):
        """a function to return all offers for a given token pair, at the depth of the direction specified

        :param token0: the address of the pay token in the offer pair
        :type token0: str
        :param token1: the address of the buy token in the offer pair
        :type token1: str
        :return: a double array of offers, the first being the offers for token0 = pay token and token1 = buy token, the second being the offers for token0 = buy token and token1 = pay token. [[[token0_amount, token_1_amount, offer_id]], [[token1_amount, token0_amount, offer_id]], depth]
        :rtype: list
        """

        # get the number of orders in the book for the given token pair
        offer_count = self.market.get_offer_count(token0, token1)

        # get the sorted orders in the book for the given token pair
        book = self.router.get_book_from_pair(token0, token1, offer_count)

        return book 

    # a function that will either create a new book or return an existing book based upon the pair 
    def get_book(self, token0, token1): 
        """a function that will either create a new book or return an existing book that is being tracked for the pair
        
        :param token0: the address of the pay token in the offer pair
        :type token0: str
        :param token1: the address of the buy token in the offer pair
        :type token1: str
        :return: a book object
        :rtype: Book
        """

        # create the pair key 
        key = str(token0) + '/' + str(token1)

        # check if the book exists
        if key in self.books:
            return self.books[key]
        else:
            book = Book(token0, token1)
            book.populate(self.get_offers(token0, token1)[0])
            self.books[key] = book
            return book
        
    def populate_book(self, token0, token1):
        """a function to populate a book with the current offers in the market

        :param token0: the address of the pay token in the offer pair
        :type token0: str
        :param token1: the address of the buy token in the offer pair
        :type token1: str
        """

        book = self.get_book(token0, token1)
        book.populate(self.get_offers(token0, token1)[0])

        return book

    ######################################################################
    # write calls
    ######################################################################

    # a function to create and pass back an ERC20 token instance
    # TODO: add support for a contract instance
    def token(self, token_address):
        """a function to create and pass back an ERC20 token instance

        :param token_address: the address of the token
        :type token_address: str
        :return: an ERC20 token instance
        :rtype: token.Token
        """

        if self.wallet and self.key:
            return contracts.helper.ERC20Signer(self.w3, token_address, self.wallet, self.key, contract=None)
        else:
            return contracts.helper.ERC20(self.w3, token_address, contract=None)

    # a function to create and pass back a market aid instance
    def aid(self, address, contract=None):
        """a function to create and pass back a market aid instance

        :param address: the address of the market aid contract
        :type address: str
        :param contract: a market aid contract instance, defaults to None
        :type contract: class, optional
        :return: a market aid instance
        :rtype: aid.Aid
        """

        if self.wallet and self.key:
            return contracts.MarketAidSigner(self.w3, address, self.wallet, self.key, contract=contract)
        else:
            return contracts.MarketAid(self.w3, address, contract=contract)


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rubi',
 'rubi.book',
 'rubi.contracts',
 'rubi.contracts.helper',
 'rubi.data',
 'rubi.data.processing',
 'rubi.data.processing.helper',
 'rubi.data.sources',
 'rubi.data.sources.helper']

package_data = \
{'': ['*'], 'rubi.contracts.helper': ['abis/*']}

install_requires = \
['attributedict>=0.3.0,<0.4.0',
 'eth-abi>=3.0.1,<4.0.0',
 'eth-tester>=0.8.0b1,<0.9.0',
 'eth-utils>=2.1.0,<3.0.0',
 'hexbytes>=0.3.0,<0.4.0',
 'py-evm>=0.6.1a1,<0.7.0',
 'pytest>=7.2.0,<8.0.0',
 'subgrounds>=1.0.3,<2.0.0',
 'web3>=6.0.0b,<7.0.0']

extras_require = \
{':extra == "docs"': ['sphinx>=5.3.0,<6.0.0']}

setup_kwargs = {
    'name': 'rubi',
    'version': '1.2.11',
    'description': 'a python SDK for the Rubicon Protocol',
    'long_description': '# rubi\nrubi is a python SDK for the Rubicon Protocol and has a variety of functionality for interacting with the protocol. documentation related to rubi and its functionality can be found [here](https://rubi.readthedocs.io/en/latest/#). \n\n### Design Goals\nthe underlying goal of the design of rubi is to decrease user friction when interacting with the protocol, and when interacting with the sdk. we want to enable the user\nas much access as possible without requiring any input from the user. for example, by breaking up the contract functionality into read and write classes, we enable the user to\nuse the sdk to read from the protocol without passing in any keys. within the data classes, we aim to provide as much information as possible to the user without requiring any \nadditional input such as api keys. when it is clearly useful, we will add higher level classes that provide additional functionality to the user by requiring additional input. \n\n### SDK Disclaimer\n\nThis codebase is in Alpha and could contain bugs or change significantly between versions. Contributing through Issues or Pull Requests is welcome!\n\n### Protocol Disclaimer\n\nPlease refer to [this](https://docs.rubicon.finance/docs/protocol/rubicon-pools/risks) for information on the risks associated to the Rubicon Protocol.',
    'author': 'denver',
    'author_email': 'denver@rubicon.finance',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
